import streamlit as st
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()

st.title("Borrowers")

if 'borrower_id' not in st.session_state:
    st.session_state.borrower_id = None
if 'borrowers_query' not in st.session_state:
    st.session_state.borrowers_query = None


@st.cache_resource
def altscore_login():
    from altscore import AltScore
    from decouple import config
    altscore = AltScore(
        client_id=config("ALTSCORE_CLIENT_ID"),
        client_secret=config("ALTSCORE_CLIENT_SECRET"),
        environment=config("ENVIRONMENT")
    )
    return altscore


altscore = altscore_login()


@st.cache_data
def load_data_models(_altscore):
    data_models_ = altscore.borrower_central.data_models.retrieve_all()
    return data_models_


data_models = load_data_models(altscore)

with st.form(key='my_form'):
    borrower_query_by = st.selectbox(
        "Search By",
        ["self", "identity"]
    )
    search = st.text_input("Search")
    submit_button = st.form_submit_button(label='Submit')


def load_borrower(borrower_id):
    st.session_state.borrower_id = borrower_id


def render_borrower(borrower_summary):
    container = st.container(border=True)
    col1, col2 = container.columns(2)
    col1.write(f"ID: {borrower_summary.id} ({borrower_summary.persona})")
    col2.button(
        "Load",
        key=borrower_summary.id,
        on_click=load_borrower,
        args=(borrower_summary.id,)
    )
    container.json(borrower_summary.dict(by_alias=True))


if submit_button:
    st.session_state.borrowers_query = altscore.borrower_central.borrowers.query_summary(
        search=search,
        by=borrower_query_by
    )
if st.session_state.borrowers_query is not None:
    for borrower in st.session_state.borrowers_query:
        render_borrower(borrower)



import streamlit as st
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()
st.title("Evaluators")


@st.cache_resource
def altscore_login():
    from altscore import AltScore
    from decouple import config
    altscore = AltScore(
        client_id=config("ALTSCORE_CLIENT_ID"),
        client_secret=config("ALTSCORE_CLIENT_SECRET"),
        environment=config("ENVIRONMENT")
    )
    return altscore


altscore = altscore_login()

evaluators = altscore.borrower_central.evaluators.retrieve_all()

st.selectbox("Evaluator", [None] + list(sorted([e.data.label for e in evaluators])))



import os
import json
from decouple import config
from pathlib import Path
import streamlit as st
from workflows.native_tasks import native_configuration


def hide_deploy_button():
    hide_deploy_button_css = """
    <style>
        /* Replace .element-class with the actual class name or ID of the deploy button */
        .stDeployButton{
            display: none;
        }
    </style>
    """
    # Inject the CSS with the markdown component
    st.markdown(hide_deploy_button_css, unsafe_allow_html=True)


def load_task_definitions():
    with open(Path(config("PROJECT_ROOT")) / "app" / "tasks" / "task_definitions.json") as f:
        custom_configuration = json.load(f)
    return native_configuration, custom_configuration


def save_task_definitions(task_definitions):
    with open(Path(config("PROJECT_ROOT")) / "app" / "tasks" / "task_definitions.json", "w") as f:
        json.dump(task_definitions, f)


def load_workflow_definition(workflow_alias: str, workflow_version: str):
    with open(Path(config(
            "PROJECT_ROOT")) / "app" / "workflows" / f"{workflow_alias}_{workflow_version}/flow_definition.json") as f:
        return json.load(f)


def save_workflow_definition(workflow_alias: str, workflow_version: str, flow_definition: dict):
    with open(Path(config(
            "PROJECT_ROOT")) / "app" / "workflows" / f"{workflow_alias}_{workflow_version}/flow_definition.json",
              "w") as f:
        json.dump(flow_definition, f)


def list_workflows():
    workflows = [f for f in os.listdir(Path(config("PROJECT_ROOT")) / "app" / "workflows") if
                 os.path.isdir(Path(config("PROJECT_ROOT")) / "app" / "workflows" / f)]

    return [
        {"alias": "_".join(workflow.split("_")[:-1]),
         "version": workflow.split("_")[-1],
         "label": workflow} for workflow in workflows
    ]



import streamlit as st

# This must go first
st.set_page_config(
    page_title="AltScore Workflow Builder",
    page_icon=":bar_chart:",
    initial_sidebar_state="expanded",
    menu_items={
        "About": "This is a workflow builder for AltScore",
    }
)
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()



import json
import streamlit as st
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()
st.title("Packages")


@st.cache_resource
def altscore_login():
    from altscore import AltScore
    from decouple import config
    altscore = AltScore(
        client_id=config("ALTSCORE_CLIENT_ID"),
        client_secret=config("ALTSCORE_CLIENT_SECRET"),
        environment=config("ENVIRONMENT")
    )
    return altscore


altscore = altscore_login()



import streamlit as st
from utils import save_workflow_definition, save_task_definitions


def add_native_task(task_name, native_task, workflow_alias, workflow_version, flow_definition):
    flow_definition["task_instances"][task_name] = {"type": native_task, "to": {}}
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)


def remove_native_task(task_name, workflow_alias, workflow_version, flow_definition):
    del flow_definition["task_instances"][task_name]
    for task, details in flow_definition["task_instances"].items():
        if 'to' in details:
            if task_name in details['to']:
                details['to'].pop(task_name)
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)
    st.success(f"Task {task_name} deleted successfully.")



import streamlit as st
from utils import save_workflow_definition


@st.cache_resource
def altscore_login():
    from altscore import AltScore
    from decouple import config
    altscore = AltScore(
        client_id=config("ALTSCORE_CLIENT_ID"),
        client_secret=config("ALTSCORE_CLIENT_SECRET"),
        environment=config("ENVIRONMENT")
    )
    return altscore


@st.cache_data
def load_sources(_altscore):
    sources_ = _altscore.borrower_central.store_sources.retrieve_all()
    return sources_


@st.cache_data
def load_data_models(_altscore):
    data_models_ = _altscore.borrower_central.data_models.retrieve_all()
    return data_models_


@st.experimental_dialog("Choose the source package")
def add_input_override(selected_task, workflow_alias, workflow_version, flow_definition):
    altscore = altscore_login()
    sources_ = load_sources(altscore)
    data_models_ = load_data_models(altscore)

    selected_sources = st.selectbox('Select Source', [source.data.id for source in sources_])
    selected_data_models = st.multiselect('Select Access key', [data_model.data.key for data_model in data_models_])
    if len(selected_sources) == 0:
        st.error("Please select at least one source")
        return

    if st.button("Add"):
        flow_definition['task_instances'][selected_task]['input_override'] = {
            f"borrower_package_{selected_sources}":
                {
                    "type": "borrower_package",
                    "borrower_id_input_alias": "borrower_id",
                    "source_id": selected_sources
                }
        }
        if len(selected_data_models) == 0:
            selected_data_models = None
        else:
            selected_data_models = f'by_{convert_snake_case_to_camel_case(selected_data_models[0])}'
            flow_definition['task_instances'][selected_task]['input_override']['package_alias'] = selected_data_models
        save_workflow_definition(workflow_alias, workflow_version, flow_definition)
        st.rerun()


def convert_snake_case_to_camel_case(snake_str):
    components = snake_str.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])



import os
from decouple import config
from pathlib import Path
import streamlit as st
from utils import save_workflow_definition, save_task_definitions


def determine_levels(task_nodes):
    levels = {}
    max_width = 1500  # Define the maximum width of the graph
    level_spacing = {}
    no_incoming = set(task_nodes.keys())

    # Find tasks with no incoming edges (start nodes)
    for task, info in task_nodes.items():
        for next_task in info.get("to", {}):
            if next_task in no_incoming:
                no_incoming.remove(next_task)

    # Assign levels using a queue
    queue = [(task, 0) for task in no_incoming]
    while queue:
        current, level = queue.pop(0)
        if current not in levels:
            levels[current] = level
        for next_task in task_nodes[current].get("to", {}):
            if next_task not in levels or level + 1 > levels[next_task]:
                levels[next_task] = level + 1
                queue.append((next_task, level + 1))

    # Calculate spacing for each level
    for level in set(levels.values()):
        level_tasks = [task for task, lvl in levels.items() if lvl == level]
        width_per_task = max_width / max(1, len(level_tasks))
        level_spacing[level] = [width_per_task * (i + 0.5) for i in range(len(level_tasks))]

    return levels, level_spacing


def add_item(task_details, key, item_details, task_definitions, selected_task):
    """Add an item (input/output) to the task details."""
    if item_details['alias']:
        task_details[key].append(item_details)
        task_definitions[selected_task] = task_details
        save_task_definitions(task_definitions)
        st.sidebar.success(f"New {key[:-1]} added successfully!")
    else:
        st.sidebar.error(f"{key[:-1].capitalize()} alias cannot be empty.")
    st.rerun()


def update_edges(action, src, tgt, workflow_alias, workflow_version, flow_definition):
    data_to_append = {src: {}}
    if action == 'add' and tgt not in flow_definition["task_instances"][src]["to"]:
        flow_definition["task_instances"][src]["to"][tgt] = {}
    elif action == 'remove' and tgt in flow_definition["task_instances"][src]["to"]:
        flow_definition["task_instances"][src]["to"].pop(tgt)
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)


def create_task(new_task_name, task_definitions, new_task_inputs, new_task_outputs, workflow_alias, workflow_version,
                flow_definition):
    new_task_name = new_task_name.replace(" ", "").lower()
    new_task_inputs = new_task_inputs.replace(" ", "").lower()
    new_task_outputs = new_task_outputs.replace(" ", "").lower()

    task_definitions[new_task_name] = {
        "name": new_task_name,
        "inputs": [{"alias": inp.strip()} for inp in new_task_inputs.split(',') if inp],
        "outputs": [{"alias": out.strip()} for out in new_task_outputs.split(',') if out]
    }
    flow_definition["task_instances"][new_task_name] = {"type": new_task_name, "to": {}}
    new_task_class = new_task_name.replace("_", " ").title().replace(" ", "")
    save_task_definitions(task_definitions)
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)
    add_custom_class(Path(config("PROJECT_ROOT")) / "app" / "tasks" / "__init__.py", new_task_name,
                     new_task_class)
    create_task_file(Path(config("PROJECT_ROOT")) / "app" / "tasks" / f"{new_task_name}.py",
                     new_task_class,
                     new_task_inputs, new_task_outputs)

    st.success("Task created successfully!")


def delete_task(selected_task, custom_task_definitions, flow_definition, workflow):
    if selected_task in custom_task_definitions:
        del custom_task_definitions[selected_task]
    if selected_task in flow_definition["task_instances"]:
        del flow_definition["task_instances"][selected_task]
    for task, details in flow_definition["task_instances"].items():
        if 'to' in details:
            if selected_task in details['to']:
                details['to'].pop(selected_task)

    save_task_definitions(custom_task_definitions)
    save_workflow_definition(workflow['alias'], workflow['version'], flow_definition)
    remove_custom_class(Path(config
                             ("PROJECT_ROOT")) / "app" / "tasks" / "__init__.py", selected_task)
    delete_task_file(Path(config("PROJECT_ROOT")) / "app" / "tasks" / f"{selected_task}.py")
    st.success(f"Task '{selected_task}' deleted successfully!")


def add_custom_class(file_path, new_task_name, new_task_class):
    new_import = f"from app.tasks.{new_task_name.lower()} import {new_task_class}\n"
    new_dict_entry = f'    "{new_task_name.lower()}": {new_task_class},\n'

    # Read the existing content of the file
    with open(file_path, 'r') as file:
        lines = file.readlines()

    # Prepare to modify the file content
    with open(file_path, 'w') as file:
        import_added = False
        dict_entry_added = False
        i = 0  # Index to keep track of the current line in the loop
        while i < len(lines):
            line = lines[i]
            # Check if the import section has ended and our import is not yet added
            if 'import' in line and not import_added:
                if 'from app.tasks' not in line:
                    file.write(new_import)
                    import_added = True

            # Check where to add the new dictionary entry
            if 'custom_functions:' in line and not dict_entry_added:
                file.write(line)
                i += 1  # Move to the next line
                if i < len(lines):
                    file.write(lines[i])  # Write the next line which starts the dictionary
                    file.write(new_dict_entry)  # Add new entry
                    dict_entry_added = True
                    i += 1  # Skip writing the next line manually
                    continue

            # Write the current line back to the file
            file.write(line)
            i += 1

        # In case imports or dictionary entries are at the end and no modification was triggered
        if not import_added:
            file.write(new_import)
        if not dict_entry_added:
            file.write(f"\ncustom_functions['{new_task_name.lower()}'] = {new_task_class}\n")


def remove_custom_class(file_path, task_name):
    # Define the target import line and dictionary entry
    target_import = f"from app.tasks.{task_name.lower()} import "
    target_dict_key = f'    "{task_name.lower()}": '

    # Read the existing content of the file
    with open(file_path, 'r') as file:
        lines = file.readlines()

    # Prepare to modify the file content
    with open(file_path, 'w') as file:
        skip_import = False
        skip_dict_entry = False
        for line in lines:
            # Check for the import line to skip
            if line.startswith(target_import):
                skip_import = True
                continue  # Skip this import line

            # Check for the dictionary entry line to skip
            if line.startswith(target_dict_key):
                skip_dict_entry = True
                continue  # Skip this dictionary entry line

            # Write back other lines
            file.write(line)

    # Confirm removal status for diagnostics
    if skip_import and skip_dict_entry:
        print(f"Removed import and dictionary entry for {task_name}.")
    else:
        print(f"Could not find import or dictionary entry for {task_name}.")


def create_task_file(file_path, class_name, inputs, outputs):
    # Try to split the inputs and outputs by ',', if not possible, use the whole string as list with one element
    inputs = inputs.split(',') if ',' in inputs else [inputs]
    outputs = outputs.split(',') if ',' in outputs else [outputs]

    content = f"""from typing import Dict
from workflows.model.task_definition import TaskConfiguration
from workflows.model.task_instance import TaskInstanceConfiguration
from workflows.model.custom_function import CustomFunction


class {class_name}(CustomFunction):
    \"\"\"
    Inputs: {inputs}
    Outputs: {outputs}
    \"\"\"

    async def _execute(self, inputs: dict, task_configuration: TaskConfiguration,
                       task_instance_configuration: TaskInstanceConfiguration, context: Dict):
        # Example processing of inputs
"""

    for input_name in inputs:
        content += f"        {input_name} = inputs.get('{input_name}')\n"

    content += "\n"

    content += "        return {\n"
    for output in outputs:
        content += f"            '{output}': {output},  # Adjust the output as needed\n"
    content += "        }\n"

    with open(file_path, 'w') as file:
        file.write(content)
    print(f"File created: {file_path}")


def delete_task_file(file_path):
    if os.path.exists(file_path):
        os.remove(file_path)
        print(f"File deleted: {file_path}")
    else:
        print(f"File not found: {file_path}")



import streamlit as st
from utils import save_workflow_definition


def add_workflow_args(new_workflow_arg, workflow_alias, workflow_version, flow_definition):
    flow_definition["workflow_args"].append({"alias": new_workflow_arg})
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)


def remove_workflow_args(workflow_arg, workflow_alias, workflow_version, flow_definition):
    flow_definition["workflow_args"] = [arg for arg in flow_definition["workflow_args"] if arg["alias"] != workflow_arg]
    save_workflow_definition(workflow_alias, workflow_version, flow_definition)



import streamlit as st


def task_instance_dropdown(workflow_definition):
    options = []
    counter = 1
    for k, v in workflow_definition["task_instances"].items():
        options.append(
            {
                "task_instance_index": counter,  # this is used to sort the options
                "task_instance_key": k,
                "task_instance_type": v["type"]
            }
        )
        counter += 1
    return st.selectbox(
        "Select Task Instance",
        options,
        format_func=lambda x: f"{x['task_instance_index']} - {x['task_instance_key']} ({x['task_instance_type']})"
    )


def task_hitters(target_task_instance: str, workflow_definitions: dict):
    hitter_instances = []
    for task_instance_name, task_instance_data in workflow_definitions["task_instances"].items():
        if target_task_instance in task_instance_data.get("to", []):
            hitter_instances.append(task_instance_name)
    return list(set(hitter_instances))


def task_instance_graph(task_instance_key, workflow_definitions):
    task_instance_data = workflow_definitions["task_instances"].get(task_instance_key)
    st.json(task_instance_data)
    if task_instance_data is None:
        return None
    col1, col2 = st.columns(2)
    # the inputs
    col1.write("Hitters")
    hitters = task_hitters(target_task_instance=task_instance_key, workflow_definitions=workflow_definitions)
    task_dict = {}
    for hitter in hitters:
        task_dict[hitter] = col1.button(hitter)
    # the outputs
    col2.write("Targets")
    for target in task_instance_data["to"]:
        task_dict[target] = col2.button(target)
    return task_dict



import subprocess
import os
from dotenv import load_dotenv
from decouple import config


def launch_streamlit_app():
    print(os.getcwd())
    load_dotenv(dotenv_path='.env', override=True)
    print(os.environ.get("PROJECT_ROOT", "N/A"))
    # Get the directory of the current script
    dir_path = os.path.dirname(os.path.realpath(__file__))
    # Path to your Streamlit app script
    app_path = os.path.join(dir_path, 'Home.py')
    # Launch the Streamlit app
    subprocess.run(['streamlit', 'run', app_path], check=True)


if __name__ == "__main__":
    launch_streamlit_app()



import json
import streamlit as st
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()
st.title("Packages")


@st.cache_resource
def altscore_login():
    from altscore import AltScore
    from decouple import config
    altscore = AltScore(
        client_id=config("ALTSCORE_CLIENT_ID"),
        client_secret=config("ALTSCORE_CLIENT_SECRET"),
        environment=config("ENVIRONMENT")
    )
    return altscore


altscore = altscore_login()


@st.cache_data
def load_sources(_altscore):
    sources_ = altscore.borrower_central.store_sources.retrieve_all()
    return sources_


sources_ = load_sources(altscore)

with st.form(key='my_form'):
    if 'borrower_id' not in st.session_state:
        st.session_state.borrower_id = None
        borrower_id = st.text_input("Borrower ID").replace('"', '').replace("'", "").strip()
    else:
        borrower_id = st.text_input("Borrower ID", value=st.session_state.borrower_id).replace('"', '').replace("'",
                                                                                                                "").strip()
    source_id = st.selectbox("Source", [None] + list(sorted([s.data.id for s in sources_])))
    submit_button = st.form_submit_button(label='Submit')


def set_package_content(package_id):
    package = altscore.borrower_central.store_packages.retrieve(package_id)
    package.get_content()
    st.session_state.package_content = json.loads(package.content)


def render_package(package):
    st.json(package.data.dict(by_alias=True))
    st.button(
        "Load Content",
        key=package.data.id,
        on_click=set_package_content,
        args=(package.data.id,)
    )


if submit_button:
    packages = altscore.borrower_central.store_packages.query(
        borrower_id=borrower_id if len(borrower_id) > 4 else None,
        source_id=source_id
    )
    for package in packages:
        render_package(package)

if 'package_content' not in st.session_state:
    st.session_state.package_content = None
if st.session_state.package_content is not None:
    st.json(st.session_state.package_content)



from setuptools import find_packages, setup

long_description = "Streamlit-based workflow builder for AltScore"

setup(
    name="altscore-workflow-builder",
    version="0.0.24",
    description="Streamlit-based workflow builder for AltScore",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/altscore/altscore-python",
    author="AltScore",
    author_email="developers@altscore.ai",
    license="MIT",
    entry_points={
        "console_scripts": [
            "run-workflow-builder = altscore_workflow_builder.launch_app:launch_streamlit_app",
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "python-dotenv",
        "loguru",
        "click",
        "requests",
        "pydantic==1.10.13",
        "httpx",
        "stringcase",
        "python-decouple",
        "python-dateutil==2.8.2",
        "pyjwt",
        "altscore",
        "streamlit",
        "streamlit-agraph",
    ],
    extras_require={},
    python_requires=">=3.8",
)



import streamlit as st


def task_graph(task_key: str, task_definitions: dict, workflow_definitions: dict):
    st.subheader(f"Task: {task_key}")
    task_type = workflow_definitions["task_instances"][task_key]["type"]
    task_instance_data = task_definitions.get(task_type)
    if task_instance_data is None:
        return None
    st.json(task_instance_data)



import streamlit as st
from altscore_workflow_builder.utils import list_workflows, load_workflow_definition, load_task_definitions
from altscore_workflow_builder.workflow import task_instance_dropdown, task_instance_graph
from altscore_workflow_builder.task import task_graph
from altscore_workflow_builder.utils import hide_deploy_button

hide_deploy_button()
st.title("Workflow Inspector")
all_workflows = list_workflows()
native_task_definitions, custom_task_definitions = load_task_definitions()
task_definitions = {**native_task_definitions, **custom_task_definitions}
workflow = st.selectbox(label="Select Workflow", options=all_workflows, format_func=lambda x: x["label"])
workflow_definition = load_workflow_definition(
    workflow_alias=workflow["alias"],
    workflow_version=workflow["version"]
)

selected_task_instance = task_instance_dropdown(workflow_definition)
task_dict = task_instance_graph(selected_task_instance["task_instance_key"], workflow_definitions=workflow_definition)
selected_task = [k for k, v in task_dict.items() if v]

if len(selected_task) > 0:
    selected_task = selected_task[0]
    task_graph(selected_task, task_definitions=task_definitions, workflow_definitions=workflow_definition)



import streamlit as st
from altscore_workflow_builder.utils import list_workflows, load_workflow_definition, load_task_definitions, \
    save_task_definitions, save_workflow_definition
from altscore_workflow_builder.custom_tasks_utils import determine_levels, add_item, update_edges, create_task, \
    delete_task
from altscore_workflow_builder.input_override_utils import add_input_override
from altscore_workflow_builder.flow_definition_utils import add_workflow_args, remove_workflow_args
from altscore_workflow_builder.native_tasks_utils import add_native_task, remove_native_task
from altscore_workflow_builder.utils import hide_deploy_button
from streamlit_agraph import agraph, Node, Edge, Config

st.set_page_config(layout="wide")
hide_deploy_button()
st.sidebar.header("Graph Configuration")
workflow = st.sidebar.selectbox("Select Workflow", list_workflows())
flow_definition = load_workflow_definition(workflow['alias'], workflow['version'])
native_task_definitions, custom_task_definitions = load_task_definitions()
task_definitions = {**native_task_definitions, **custom_task_definitions}

# Create columns for the buttons
col1, col2, col3, col4 = st.columns(4)

# UI for creating custom task
if 'create_task' not in st.session_state:
    st.session_state['create_task'] = False
with col1:
    if st.button('Create Custom Task'):
        st.session_state.create_task = not st.session_state.create_task
if st.session_state.create_task:
    with st.form("create_task_form"):
        new_task_name = st.text_input("Task Name")
        new_task_inputs = st.text_area("Inputs (comma-separated)")
        new_task_outputs = st.text_area("Outputs (comma-separated)")
        submitted = st.form_submit_button("Submit Task")
        if submitted:
            if new_task_name and new_task_name not in custom_task_definitions:
                create_task(new_task_name, custom_task_definitions, new_task_inputs, new_task_outputs,
                            workflow['alias'],
                            workflow['version'], flow_definition)
                st.session_state.create_task = False
                st.rerun()
            else:
                st.error("Task name is required or already exists.")

# UI for adding native task
if 'add_native_task' not in st.session_state:
    st.session_state['add_native_task'] = None
with col2:
    if st.button('Create Native Task'):
        st.session_state.add_native_task = not st.session_state.add_native_task

if st.session_state.add_native_task:
    with st.form("create_native_task_form"):
        new_task_name = st.text_input("Task Name")
        native_task = st.selectbox("Native Task", list(native_task_definitions.keys()))
        if st.form_submit_button("Submit Task"):
            if new_task_name and new_task_name not in custom_task_definitions:
                add_native_task(new_task_name, native_task, workflow['alias'], workflow['version'], flow_definition)
                st.session_state.add_native_task = False
                st.rerun()
            else:
                st.error("Task name is required or already exists.")

# UI for workflow arguments
st.sidebar.title("Workflow Arguments")
st.sidebar.json(flow_definition.get("workflow_args", []))
alias_name = st.sidebar.text_input("Alias for Workflow Argument", key="workflow_arg_alias")
if st.sidebar.button("Add new Workflow Argument"):
    add_workflow_args(alias_name, workflow['alias'], workflow['version'], flow_definition)
    st.sidebar.success(f"New workflow argument added successfully!")
    st.rerun()
if st.sidebar.button("Remove selected Workflow Argument"):
    remove_workflow_args(alias_name, workflow['alias'], workflow['version'], flow_definition)
    st.sidebar.success("Workflow argument removed successfully!")
    st.rerun()

# UI for the graph
nodes = []
edges = []
task_nodes = flow_definition["task_instances"]
levels, level_spacing = determine_levels(task_nodes)
all_task_names = list(task_nodes.keys())

for task_name, task_info in flow_definition["task_instances"].items():
    inputs = ", ".join([inp['alias'] for inp in task_definitions.get(task_info['type'], {}).get('inputs', [])])
    outputs = ", ".join([out['alias'] for out in task_definitions.get(task_info['type'], {}).get('outputs', [])])
    input_overrides = ", ".join(
        [inp for inp in flow_definition['task_instances'][task_name].get('input_overrides', {})])
    label = f"{task_name}"
    tooltip = f"Inputs: {inputs}\nOutputs: {outputs}\nInput Overrides: {input_overrides}"
    level = levels[task_name]
    task_category = "Custom" if task_name in custom_task_definitions else "Native"
    shape = "dot" if task_category == "Custom" else "diamond"
    color = "lightblue" if task_category == "Custom" else "lightgreen"
    nodes.append(
        Node(id=task_name, label=label, shape=shape, color=color, size=50, x=level_spacing[level].pop(0),
             y=level * 150,
             title=tooltip))

    for next_task in task_info.get("to", []):
        edges.append(Edge(source=task_name, target=next_task, type="STRAIGHT"))

# Configuration for agraph
config = Config(
    height=900,
    width='100%',
    directed=True,
    nodeHighlightBehavior=True,
    highlightColor='#88c999',
    collapsible=True,
    node={'labelProperty': 'label', 'font_size': 20},
    link={'labelProperty': 'label', 'renderLabel': True},
    staticGraph=True,
    physics={
        "solver": 'barnesHut',
        "hierarchical": False
    }
)

# Display the graph
selection = agraph(nodes=nodes, edges=edges, config=config)

if selection:
    task_info = task_nodes[selection]
    selected_task = st.sidebar.selectbox("Select Task", all_task_names, index=all_task_names.index(selection))
    if selected_task:
        task_info = task_nodes[selected_task]
        task_details = task_definitions.get(task_info['type'], {})

        # UI for add/remove Inputs, Outputs
        for detail_key in ['inputs', 'outputs']:
            st.sidebar.title(f"{detail_key.capitalize()} Management")
            st.sidebar.json(task_details.get(detail_key, []))

            alias_name = st.sidebar.text_input(f"Alias for {detail_key[:-1]}", key=f"{detail_key}_alias")
            item_details = {"alias": alias_name}
            button_label = f"Add new {detail_key[:-1]}"
            if st.sidebar.button(button_label):
                if selected_task in list(native_task_definitions.keys()):
                    st.sidebar.error(f"Cannot add {detail_key} to a native task.")
                else:
                    add_item(task_details, detail_key, item_details, custom_task_definitions, selected_task)

            aliases = [item['alias'] if 'alias' in item else f"{item['key']}:{item['value']}" for item in
                       task_details.get(detail_key, [])]
            item_to_remove = st.sidebar.selectbox(f"Select {detail_key[:-1]} to remove", aliases,
                                                  key=f"{detail_key}_remove")
            if st.sidebar.button(f"Add borrower_package to {detail_key[:-1]}"):
                add_input_override(selected_task, workflow['alias'], workflow['version'], flow_definition)

            if st.sidebar.button(f"Remove selected {detail_key[:-1]}"):
                if selected_task in list(native_task_definitions.keys()):
                    st.sidebar.error(f"Cannot remove {detail_key} from a native task.")
                else:
                    task_details[detail_key] = [item for item in task_details[detail_key] if
                                                item['alias'] != item_to_remove]
                    task_definitions[selected_task] = task_details
                    save_task_definitions(custom_task_definitions)
                    st.sidebar.success(f"{detail_key[:-1].capitalize()} removed successfully!")
                    st.rerun()

        # UI for Delete Task
        if 'confirm_delete' not in st.session_state:
            st.session_state['confirm_delete'] = None
        with col4:
            if st.button('Delete Task', key="delete_task", type="primary"):
                if st.session_state.confirm_delete == selected_task:
                    if selected_task in custom_task_definitions:
                        delete_task(selected_task, custom_task_definitions, flow_definition, workflow)
                        st.session_state.confirm_delete = None  # Reset the confirmation state
                        st.rerun()
                    else:
                        remove_native_task(selected_task, workflow['alias'], workflow['version'], flow_definition)
                        st.session_state.confirm_delete = None
                        st.rerun()
                else:
                    st.session_state.confirm_delete = selected_task
                    st.warning(
                        f"Are you sure you want to delete '{selected_task}'? Click 'Delete Task' again to confirm.")

    # UI for edge management
    st.sidebar.title("Manage Edges")
    source_task = st.sidebar.selectbox("Source Task", all_task_names, index=all_task_names.index(selection))
    target_task = st.sidebar.selectbox("Target Task", all_task_names, index=all_task_names.index(selection))
    if st.sidebar.button("Add Edge"):
        update_edges('add', source_task, target_task, workflow['alias'], workflow['version'], flow_definition)
        st.rerun()
    if st.sidebar.button("Remove Edge"):
        update_edges('remove', source_task, target_task, workflow['alias'], workflow['version'], flow_definition)
        st.rerun()
